
/*
** (c) 2003 Herbert Poetzl
**
** V0.01	test for syscall
** V0.02	spawn arbitrary process
** V0.03	destroy virtual
** V0.04	clone & pivot root
** V0.05	release cleanups
**
*/

#define VERSION "V0.05"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <syscall.h>
#include <asm/unistd.h>
#include <unistd.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/mount.h>

#ifndef CLONE_NEWNS
#define CLONE_NEWNS 0x00020000
#endif

 
#include <linux/virtual.h>

_syscall2(int, pivot_root, const char *, new_root, const char *, put_old);
       
int pivot_root(const char *new_root, const char *put_old);


static	char	err_flag = 0;

static	char	opt_create = 0;
static	char	opt_migrate = 0;
static	char	opt_number = 0;
static	char	opt_time = 0;
static	char	opt_exec = 0;
static	char	opt_fork = 0;
static	char	opt_pivot = 0;

static	int	number = 1;
static	int	flags = 0;
static	int	sleep_time = 0;


static	char	oldpath[] = "/_old"; 

static 	char **	exec_args = NULL;
static 	char *	exec_file = NULL;
static 	char *	new_root = NULL;
static 	char *	put_old = NULL;
static	char *	cmd_name = NULL;


_syscall2(int, create_virtual, int, vid, struct virtual_setup *, setup);
_syscall1(int, migrate, int, vid);

static	void *	child_stack;


int 	child(void *p) {
	int r;
	
        setenv("PS1", "@@ ", 1);
	
	if (opt_pivot) {
	    int len = strlen(new_root);
	
	    put_old = calloc(len+sizeof(oldpath), sizeof(char));
	    if (put_old) {
	    	snprintf(put_old, len+sizeof(oldpath)+1, 
			"%s%s", new_root, oldpath);

	    	r = pivot_root(new_root, put_old);
	    	if (r)
		    perror("pivot");
		
		r = chdir("/");
	    	if (r)
		    perror("chdir");

		if (opt_migrate) {
		    r = migrate(number);
		    if (r)
		    	perror("migrate");
		}
				    
	    } else {
	    	perror("calloc");
	    }
	}
	
	execvp(exec_file, exec_args);
        perror("execvp");
        return 1;
}

int	main(int argc, char *argv[])
{
        extern int optind;
        extern char *optarg;
        char c, errflg = 0;
        pid_t pid, p;
	int r;


        cmd_name = argv[0];
        while ((c = getopt(argc, argv, "hcmE:F:N:O:P:T:")) != EOF) {
            switch (c) {
            case 'h':
                fprintf(stderr,
                    "This is %s " VERSION "\n"
                    "options are:\n"
                    "-h        print this help message\n"
                    "-c        create virtual\n"
                    "-m        migrate task virtual\n"
                    "-E <arg>  exec file <arg>\n"
                    "-F <arg>  fork (clone) file <arg>\n"
		    "-N <arg>  virtual id <arg>\n"
		    "-O <arg>  optional flags <arg>\n"
		    "-P <arg>  pivot root path <arg>\n"
		    "-T <arg>  some time (in seconds) <arg>\n",
                    cmd_name);
                exit(0);
                break;
            case 'c':
                opt_create = 1;
                break;
            case 'm':
                opt_migrate = 1;
                break;
            case 'N':
                opt_number = 1;
                number = atol(optarg);
                break;
            case 'E':
                opt_exec = 1;
                exec_file = optarg;
                break;
            case 'F':
                opt_fork = 1;
                exec_file = optarg;
                break;
            case 'O':
                flags = atol(optarg);
                break;
            case 'P':
                opt_pivot = 1;
                new_root = optarg;
                break;
            case 'T':
                opt_time = 1;
                sleep_time = atol(optarg);
                break;
            case '?':
            default:
                errflg++;
                break;
            }
        }
        if (errflg) {
            fprintf(stderr, 
                "Usage: %s -[hcmE:F:N:O:P:T:]\n"
                "%s -h for help.\n",
                cmd_name, cmd_name);
            exit(2);
        }

	if (opt_create) {
	    struct virtual_setup setup = { 0 };
	    
	    setup.flags = flags;
	    r = create_virtual(number, &setup);
	    if (r)
	    	perror("create");
	}
	if (opt_migrate && !opt_fork) {
	    r = migrate(number);
	    if (r)
	    	perror("migrate");
	}
	if (opt_time) {
	    sleep(sleep_time);
	}
	
	exec_args = &argv[optind];
	child_stack = malloc(8192);
	if (!child_stack) 
	    perror("malloc");
	child_stack += (8192-32);
	
	if (opt_fork) {
 	    pid = clone(child, child_stack, CLONE_NEWNS | SIGCHLD, NULL);
 	    if ((int) pid == -1) {
 		perror("clone");
 		exit(1);
 	    }
 	    p = waitpid(pid, NULL, 0);
 	    if ((int) p == -1) {
 		perror("waitpid");
 		exit(2);
 	    }
	} else {
	    if (opt_exec) {
	    	execvp(exec_file, exec_args);
	    	exit(1);
	    }
	}

	exit((err_flag)?1:0);
}

